"""Switcher water heater unofficial API and bridge, Network protocols.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

# fmt: off
from asyncio import (
    AbstractEventLoop, BaseTransport, DatagramProtocol, Event, Future, Queue,
    QueueEmpty, QueueFull, Transport, ensure_future)
from datetime import datetime
from functools import partial
from typing import Optional, Tuple, Union, cast

from .bridge.messages import SwitcherV2BroadcastMSG
from .devices import SwitcherV2Device

# fmt: on


class SwitcherV2UdpProtocolFactory(DatagramProtocol):
    """Represntation of the Asyncio UDP protocol factory.

    Args:
      loop: the event loop for the factory to run in.
      phone_id: the phone id of the desired device.
      device_id: the id of the desired device.
      device_password: the password of the desired device.
      queue: a ``asyncio.Queue`` for the factory to save messages in.
      run_factory_evt ``asyncio.Event`` for signaling the factory to run.

    Todo:
      * replace ``factory_future`` attribute with ``get_factory_future``
        method.

    """

    def __init__(
        self,
        loop: AbstractEventLoop,
        phone_id: str,
        device_id: str,
        device_password: str,
        queue: Queue,
        run_factory_evt: Event,
    ) -> None:
        """Initialize the protocol."""
        self._loop = loop
        self._phone_id = phone_id
        self._device_id = device_id
        self._device_password = device_password
        self._queue = queue
        self._run_factory = run_factory_evt
        self._factory_future = self._loop.create_future()
        self._factory_future.add_done_callback(self.close_transport)
        self._accept_datagrams = Event()
        self.transport = None  # type: Optional[Transport]
        self._device = None  # type: Optional[SwitcherV2Device]

    def connection_made(self, transport: BaseTransport) -> None:
        """Call on connection established."""
        self.transport = cast(Transport, transport)
        self._accept_datagrams.set()

    def datagram_received(self, data: Union[bytes, str], addr: Tuple) -> None:
        """Call on datagram recieved."""
        if self._run_factory.is_set() and self._accept_datagrams.is_set():
            ensure_future(
                self.handle_incoming_messages(data, addr), loop=self._loop
            )

    def error_received(self, exc: Optional[Exception]) -> None:
        """Call on exception recieved."""
        if exc:
            self.factory_future.set_exception(exc)
        else:
            self.factory_future.set_result(None)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """Call on connection lost."""
        self.factory_future.set_result(exc if exc else None)

    def close_transport(self, future: Future) -> None:
        """Call for closing the transport."""
        if self.transport:
            self.transport.close()

    async def handle_incoming_messages(
        self, data: Union[bytes, str], addr: Tuple
    ) -> None:
        """Use for Handling incoming messages."""
        self._accept_datagrams.clear()
        msg = SwitcherV2BroadcastMSG(self._loop, data)
        msg.init_future.add_done_callback(
            partial(self.get_device_from_message, addr[0])
        )

        return None

    def get_device_from_message(self, ip_addr: str, future: Future) -> None:
        """Use for extracting the device from the broadcast message."""
        msg = future.result()
        if msg.verified:
            if self._device_id == msg.device_id:
                if self._device:
                    # Update known device
                    self._device.update_device_data(
                        ip_addr,
                        msg.name,
                        msg.device_state,
                        msg.remaining_time_to_off,
                        msg.auto_off_set,
                        msg.power,
                        msg.current,
                        (
                            self._device.last_state_change
                            if msg.device_state == self._device.state
                            else datetime.now()
                        ),
                    )
                else:
                    # New device discoverd
                    self._device = SwitcherV2Device(
                        msg.device_id,
                        ip_addr,
                        msg.mac_address,
                        msg.name,
                        msg.device_state,
                        msg.remaining_time_to_off,
                        msg.auto_off_set,
                        msg.power,
                        msg.current,
                        self._phone_id.lower(),
                        self._device_password.lower(),
                        datetime.now(),
                    )
                try:
                    self._queue.put_nowait(self._device)
                except QueueFull:
                    try:
                        self._queue.get_nowait()
                    except QueueEmpty:
                        pass
                    ensure_future(
                        self._queue.put(self._device), loop=self._loop
                    )
        self._accept_datagrams.set()

    @property
    def factory_future(self) -> Future:
        """asyncio.Future: Representing the initialization status."""
        return self._factory_future



"""Switcher water heater unofficial API and bridge, Global constants.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

ENCODING_CODEC = "utf-8"
REMOTE_SESSION_ID = "00000000"
REMOTE_KEY = b"00000000000000000000000000000000"
SOCKET_PORT = 9957
SOCKET_BIND_TUP = ("0.0.0.0", 20002)  # nosec
STATE_RESPONSE_ON = "0100"
STATE_ON = "on"
STATE_RESPONSE_OFF = "0000"
STATE_OFF = "off"
STATE_UNKNOWN = "unknown"
COMMAND_ON = "1"
COMMAND_OFF = "0"
NO_TIMER_REQUESTED = "00000000"
ENABLE_SCHEDULE = "01"
DISABLE_SCHEDULE = "00"

SUNDAY = "Sunday"
MONDAY = "Monday"
TUESDAY = "Tuesday"
WEDNESDAY = "Wednesday"
THURSDAY = "Thursday"
FRIDAY = "Friday"
SATURDAY = "Saturday"
ALL_DAYS = "Every day"

WEEKDAY_TUP = (MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY)

HEX_TO_DAY_DICT = {
    0x02: MONDAY,
    0x04: TUESDAY,
    0x08: WEDNESDAY,
    0x10: THURSDAY,
    0x20: FRIDAY,
    0x40: SATURDAY,
    0x80: SUNDAY,
}

DAY_TO_INT_DICT = {
    MONDAY: 2,
    TUESDAY: 4,
    WEDNESDAY: 8,
    THURSDAY: 16,
    FRIDAY: 32,
    SATURDAY: 64,
    SUNDAY: 128,
}

# weekdays, start-time, end-time
SCHEDULE_CREATE_DATA_FORMAT = "01{}01{}{}"

WAITING_TEXT = "waiting_for_data"

SCHEDULE_DUE_TODAY_FORMAT = "Due today at {}"
SCHEDULE_DUE_TOMMOROW_FORMAT = "Due tommorow at {}"
SCHEDULE_DUE_ANOTHER_DAY_FORMAT = "Due next {} at {}"

STRUCT_PACKING_FORMAT = "<I"  # little-endian unsigned int

HANDLED_EXCEPTIONS = (
    AttributeError,
    FloatingPointError,
    IndexError,
    LookupError,
    OverflowError,
    TypeError,
    UnicodeDecodeError,
    ValueError,
    ZeroDivisionError,
)



"""Switcher Packet Response Messages."""

# fmt: off
from asyncio import AbstractEventLoop, Future, ensure_future
from binascii import hexlify
from enum import Enum
from typing import List

from ..consts import (ENCODING_CODEC, HANDLED_EXCEPTIONS, STATE_OFF, STATE_ON,
                      STATE_RESPONSE_OFF, STATE_RESPONSE_ON, STATE_UNKNOWN)
from ..errors import DecodingError
from ..schedules import SwitcherV2Schedule
from ..tools import convert_seconds_to_iso_time

# fmt: on


ResponseMessageType = Enum(
    "ResponseMessageType",
    [
        "AUTO_OFF",
        "CONTROL",
        "CREATE_SCHEDULE",
        "DELETE_SCHEDULE",
        "DISABLE_ENABLE_SCHEDULE",
        "GET_SCHEDULES",
        "LOGIN",
        "STATE",
        "UPDATE_NAME",
    ],
)


class SwitcherV2BaseResponseMSG:
    """Represntation of the switcher v2 base response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.
      msg_type: the message type as described in the ''ResponseMessageType''
        Enum class.

    """

    def __init__(
        self,
        loop: AbstractEventLoop,
        response: bytes,
        msg_type: ResponseMessageType,
    ) -> None:
        """Initialize the base object."""
        self._loop = loop
        self._unparsed_response = response
        self._msg_type = msg_type

    @property
    def unparsed_response(self) -> bytes:
        """bytes: Return The raw response from the device."""
        return self._unparsed_response

    @property
    def successful(self) -> bool:
        """bool: Indicating whether or not the request was successful."""
        return self._unparsed_response is not None

    @property
    def msg_type(self) -> ResponseMessageType:
        """aioswitcher.api.messages.ResponseMessageType: the message type."""
        return self._msg_type


class SwitcherV2LoginResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 login response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the login response."""
        super().__init__(loop, response, ResponseMessageType.LOGIN)
        try:
            self._session_id = hexlify(response)[16:24].decode(ENCODING_CODEC)
        except HANDLED_EXCEPTIONS as exc:
            raise DecodingError(
                "failed to parse login response message"
            ) from exc

    @property
    def session_id(self) -> str:
        """str: Return the retrieved session id."""
        return self._session_id


class SwitcherV2StateResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 state response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    Todo:
      * replace ``init_future`` attribute with ``get_init_future`` method.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the state response message."""
        super().__init__(loop, response, ResponseMessageType.STATE)
        self._power_consumption = 0
        self._electric_current = 0.0
        self._init_future = loop.create_future()
        ensure_future(self.initialize(response), loop=loop)

    async def initialize(self, response: bytes) -> None:
        """Finish the initialization of the message and update the future object.

        Args:
          response: the raw response from the device.

        """
        try:
            temp_power = hexlify(response)[154:162]
            self._power_consumption = int(
                temp_power[2:4] + temp_power[0:2], 16
            )
            self._electric_current = round(
                (self._power_consumption / float(220)), 1
            )

            temp_time_left = hexlify(response)[178:186]
            temp_time_left_seconds = int(
                temp_time_left[6:8]
                + temp_time_left[4:6]
                + temp_time_left[2:4]
                + temp_time_left[0:2],
                16,
            )
            self._time_to_auto_off = await convert_seconds_to_iso_time(
                self._loop, temp_time_left_seconds
            )

            temp_auto_off = hexlify(response)[194:202]
            temp_auto_off_seconds = int(
                temp_auto_off[6:8]
                + temp_auto_off[4:6]
                + temp_auto_off[2:4]
                + temp_auto_off[0:2],
                16,
            )
            self._auto_off_set = await convert_seconds_to_iso_time(
                self._loop, temp_auto_off_seconds
            )

            temp_state = hexlify(response)[150:154].decode(ENCODING_CODEC)
            self._state = (
                STATE_ON
                if temp_state == STATE_RESPONSE_ON
                else STATE_OFF
                if temp_state == STATE_RESPONSE_OFF
                else STATE_UNKNOWN
            )

            self.init_future.set_result(self)
        except HANDLED_EXCEPTIONS as exc:
            self.init_future.set_exception(exc)

        return None

    @property
    def state(self) -> str:
        """str: Return the state."""
        return self._state

    @property
    def time_left(self) -> str:
        """str: Return the time left to auto-off."""
        return self._time_to_auto_off

    @property
    def auto_off(self) -> str:
        """str: Return the auto-off configuration value."""
        return self._auto_off_set

    @property
    def power(self) -> int:
        """int: Return the current power consumption in watts."""
        return self._power_consumption

    @property
    def current(self) -> float:
        """float: Return the power consumption in amps."""
        return self._electric_current

    @property
    def init_future(self) -> Future:
        """asyncio.Future: Return the future of the initialization."""
        return self._init_future


class SwitcherV2ControlResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 control response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the Control response message."""
        super().__init__(loop, response, ResponseMessageType.CONTROL)


class SwitcherV2SetAutoOffResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 set auto-off response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the Set Auto-Off response message."""
        super().__init__(loop, response, ResponseMessageType.AUTO_OFF)


class SwitcherV2UpdateNameResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 update name response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the Set Name response message."""
        super().__init__(loop, response, ResponseMessageType.UPDATE_NAME)


class SwitcherV2GetScheduleResponseMSG(SwitcherV2BaseResponseMSG):
    """represnation of the switcher v2 get schedule message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    Todo:
      * the ``get_schedules`` attribute should be a method.
      * ``schdule_detais`` is ``__init__`` is yielding List[str] instead of
        List[bytes], that's not supposed to happen.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the Set Name response message."""
        super().__init__(loop, response, ResponseMessageType.GET_SCHEDULES)
        self._schedule_list = []  # type: List[SwitcherV2Schedule]

        res = hexlify(response)
        idx = res[90:-8].decode(ENCODING_CODEC)
        schedules_details = [
            idx[i : i + 32] for i in range(0, len(idx), 32)  # noqa E203
        ]

        if schedules_details:
            for i in range(len(schedules_details)):
                schedule = SwitcherV2Schedule(
                    loop, i, schedules_details  # type: ignore
                )
                self._schedule_list.append(schedule)

    @property
    def found_schedules(self) -> bool:
        """bool: Return true if found schedules in the response."""
        return self._schedule_list != []

    @property
    def get_schedules(self) -> List[SwitcherV2Schedule]:
        """list(aioswitcher.schedules.SwitcherV2Schedule): Return schedules."""
        return self._schedule_list


class SwitcherV2DisableEnableScheduleResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 dis/en schedule response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the switcher v2 dis/en schedule response message."""
        super().__init__(
            loop, response, ResponseMessageType.DISABLE_ENABLE_SCHEDULE
        )


class SwitcherV2DeleteScheduleResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 delete schedule response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the switcher v2 delete schedule response message."""
        super().__init__(loop, response, ResponseMessageType.DELETE_SCHEDULE)


class SwitcherV2CreateScheduleResponseMSG(SwitcherV2BaseResponseMSG):
    """Represntation of the switcher v2 create schedule response message.

    Args:
      loop: the event loop to perform actions in.
      response: the raw response from the device.

    """

    def __init__(self, loop: AbstractEventLoop, response: bytes) -> None:
        """Initialize the switcher v2 create schedule response message."""
        super().__init__(loop, response, ResponseMessageType.CREATE_SCHEDULE)



"""Switcher water heater unofficial API and bridge, Tools and Helpers.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

# fmt: off
from asyncio import AbstractEventLoop
from binascii import crc_hqx, hexlify, unhexlify
from datetime import time as datetime_time
from datetime import timedelta
from struct import pack
from time import localtime, mktime, strftime, strptime
from time import time as time_time
from typing import List

from .consts import (ENCODING_CODEC, HANDLED_EXCEPTIONS, HEX_TO_DAY_DICT,
                     REMOTE_KEY, STRUCT_PACKING_FORMAT)
from .errors import CalculationError, DecodingError, EncodingError

# fmt: on


def _convert_seconds_to_iso_time(all_seconds: int) -> str:
    """Convert seconds to iso time.

    Args:
      all_seconds: the total number of seconds to convert.

    Return:
      A string represnting the converted iso time in %H:%M:%S format.
      e.g. "02:24:37".

    Raises:
      aioswitcher.erros.CalculationError: when failed to convert the argument.

    Note:
      This is a private function containing blocking code.
      Please consider using ``convert_seconds_to_iso_time`` (without the `_`),
      to schedule as a task in the event loop.

    """
    try:
        minutes, seconds = divmod(int(all_seconds), 60)
        hours, minutes = divmod(minutes, 60)

        return datetime_time(
            hour=hours, minute=minutes, second=seconds
        ).isoformat()
    except HANDLED_EXCEPTIONS as ex:
        raise CalculationError("failed to convert seconds to iso time") from ex


async def convert_seconds_to_iso_time(
    loop: AbstractEventLoop, all_seconds: int
) -> str:
    """Asynchronous wrapper for _convert_seconds_to_iso_time.

    Use as async wrapper for calling _convert_seconds_to_iso_time,
    calculating the next runtime of the schedule.

    Args:
      loop: the event loop to execute the function in.
      all_seconds: the total number of seconds to convert.

    Returns:
      A string represnting the converted iso time in %H:%M:%S format.
      e.g. "02:24:37".

    """
    return await loop.run_in_executor(
        None, _convert_seconds_to_iso_time, all_seconds
    )


def _crc_sign_full_packet_com_key(data: str) -> str:
    """Calculate the crc for packets before send.

    Args:
      data: packet data to sign.

    Return:
      The calculated and signed packet data.

    Raises:
      aioswitcher.erros.EncodingError: when failed to sign the packet.

    Note:
      This is a private function containing blocking code.
      Please consider using ``crc_sign_full_packet_com_key``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        crc = hexlify(pack(">I", crc_hqx(unhexlify(data), 0x1021))).decode(
            ENCODING_CODEC
        )
        data = data + crc[6:8] + crc[4:6]
        crc = (
            crc[6:8] + crc[4:6] + (hexlify(REMOTE_KEY)).decode(ENCODING_CODEC)
        )
        crc = hexlify(pack(">I", crc_hqx(unhexlify(crc), 0x1021))).decode(
            ENCODING_CODEC
        )
        data = data + crc[6:8] + crc[4:6]

        return data
    except HANDLED_EXCEPTIONS as ex:
        raise EncodingError("failed to sign crc") from ex


async def crc_sign_full_packet_com_key(
    loop: AbstractEventLoop, data: str
) -> str:
    """Asynchronous wrapper for _crc_sign_full_packet_com_key.

    Use as async wrapper for calling _crc_sign_full_packet_com_key,
    performing crc sign to the packet data.

    Args:
      data: packet data to sign.

    Return:
      The calculated and signed packet data.

    """
    return await loop.run_in_executor(
        None, _crc_sign_full_packet_com_key, data
    )


def _convert_minutes_to_timer(minutes: str) -> str:
    """Convert on-timer minutes to hexadecimal before sending.

    Args:
      minutes: on-timer minutes to convert.

    Return:
      Hexadecimal represntation of the mintues argument.

    Raises:
      aioswitcher.erros.EncodingError: when failed hexlify.

    Note:
      This is a private function containing blocking code.
      Please consider using ``convert_minutes_to_timer``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        return hexlify(pack(STRUCT_PACKING_FORMAT, int(minutes) * 60)).decode(
            ENCODING_CODEC
        )
    except HANDLED_EXCEPTIONS as ex:
        raise EncodingError(
            "failed to create timer from {} minutes".format(str(minutes))
        ) from ex


async def convert_minutes_to_timer(
    loop: AbstractEventLoop, minutes: str
) -> str:
    """Asynchronous wrapper for _convert_minutes_to_timer.

    Use as async wrapper for calling _convert_minutes_to_timer,
    converting on-timer minutes to hexadecimal.

    Args:
      minutes: on-timer minutes to convert.

    Return:
      Hexadecimal represntation of the mintues argument.

    """
    return await loop.run_in_executor(None, _convert_minutes_to_timer, minutes)


def _convert_timedelta_to_auto_off(full_time: timedelta) -> str:
    """Convert timedelta object for auto-shutdown to hexadecimal.

    Args:
      full_time: timedelta object represnting the auto-shutdown time.

    Return:
      Hexadecimal represntation of the full_time argument.

    Raises:
      aioswitcher.erros.EncodingError: when failed hexlify.

    Note:
      This is a private function containing blocking code.
      Please consider using ``convert_timedelta_to_auto_off``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        minutes = full_time.total_seconds() / 60
        hours, minutes = divmod(minutes, 60)
        seconds = int(hours) * 3600 + int(minutes) * 60
        if 3599 < seconds < 86341:
            return hexlify(pack(STRUCT_PACKING_FORMAT, int(seconds))).decode(
                ENCODING_CODEC
            )
        raise ValueError("can only handle 1 to 24 hours on auto-off set")
    except HANDLED_EXCEPTIONS as ex:
        raise EncodingError(
            "failed to create auto-off from {} timedelta".format(
                str(full_time)
            )
        ) from ex


async def convert_timedelta_to_auto_off(
    loop: AbstractEventLoop, full_time: timedelta
) -> str:
    """Asynchronous wrapper for _convert_timedelta_to_auto_off.

    Use as async wrapper for calling _convert_timedelta_to_auto_off,
    converting timedelta auto-shutdown configuration to hexadecimal.

    Args:
      full_time: timedelta object represnting the auto-shutdown time.

    Return:
      Hexadecimal represntation of the full_time argument.

    """
    return await loop.run_in_executor(
        None, _convert_timedelta_to_auto_off, full_time
    )


def _convert_string_to_device_name(name: str) -> str:
    """Convert string device name to hexadecimal before sending.

    Args:
      name: the desired name for conerting.

    Return:
      Hexadecimal represntation of the name argument.

    Raises:
      aioswitcher.erros.EncodingError: when failed hexlify.

    Note:
      This is a private function containing blocking code.
      Please consider using ``convert_string_to_device_name``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        if 1 < len(name) < 33:
            return (
                hexlify(name.encode(ENCODING_CODEC))
                + ((32 - len(name)) * "00").encode(ENCODING_CODEC)
            ).decode(ENCODING_CODEC)
        raise ValueError("name length can vary from 2 to 32.")
    except HANDLED_EXCEPTIONS as ex:
        raise EncodingError(
            "failed to convert " + str(name) + " to device name"
        ) from ex


async def convert_string_to_device_name(
    loop: AbstractEventLoop, name: str
) -> str:
    """Asynchronous wrapper for _convert_string_to_device_name.

    Use as async wrapper for calling _convert_string_to_device_name,
    converting name to hexadecimal.

    Args:
      name: the desired name for conerting.

    Return:
      Hexadecimal represntation of the name argument.

    """
    return await loop.run_in_executor(
        None, _convert_string_to_device_name, name
    )


def _get_days_list_from_bytes(data: int) -> List[str]:
    """Extract week days from shcedule bytes data.

    Args:
      data: bytes representing the days list.

    Return:
      List of string represntation the week days included in the list.
      See ``aioswitcher.consts`` for days literals.

    Raises:
      aioswitcher.erros.DecodingError: when failed to analyze the argument.

    Note:
      This is a private function containing blocking code.
      Please consider using ``get_days_list_from_bytes``
      (without the `_`), to schedule as a task in the event loop.

    """
    days_list = []
    try:
        for day in HEX_TO_DAY_DICT:
            if day & data != 0x00:
                days_list.append(HEX_TO_DAY_DICT[day])

        return days_list
    except HANDLED_EXCEPTIONS as ex:
        raise DecodingError(
            "failed to extract week days from schedule"
        ) from ex


async def get_days_list_from_bytes(
    loop: AbstractEventLoop, data: int
) -> List[str]:
    """Asynchronous wrapper for _get_days_list_from_bytes.

    Use as async wrapper for calling _get_days_list_from_bytes,
    extracting week days from shcedule bytes data.

    Args:
      data: bytes representing the days list.

    Return:
      List of string represntation the week days included in the list.
      See ``aioswitcher.consts`` for days literals.

    """
    return await loop.run_in_executor(None, _get_days_list_from_bytes, data)


def _get_time_from_bytes(data: bytes) -> str:
    """Extract start/end time from schedule bytes.

    Args:
      data: bytes representing the start or the end time for the schedule.

    Return:
      Time string in %H:%M format.
      e.g. "20:30".

    Raises:
      aioswitcher.erros.DecodingError: when failed to analyze the argument.

    Note:
      This is a private function containing blocking code.
      Please consider using ``get_time_from_bytes``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        timestamp = int(data[6:8] + data[4:6] + data[2:4] + data[0:2], 16)

        return strftime("%H:%M", localtime(timestamp))
    except HANDLED_EXCEPTIONS as ex:
        raise DecodingError("failed to extract time from schedule") from ex


async def get_time_from_bytes(loop: AbstractEventLoop, data: bytes) -> str:
    """Asynchronous wrapper for _get_time_from_bytes.

    Use as async wrapper for calling _get_time_from_bytes,
    extracting start/end time from schedule bytes data.

    Args:
      data: bytes representing the start or the end time for the schedule.

    Return:
      Time string in %H:%M format.
      e.g. "20:30".

    """
    return await loop.run_in_executor(None, _get_time_from_bytes, data)


def _get_timestamp() -> str:
    """Generate hexadecimal represntation of the current timestamp.

    Return:
      Hexadecimal represntation of the current unix time retrieved by
      ``time.time``.

    Raises:
      aioswitcher.erros.DecodingError: when failed to analyze the timestamp.

    Note:
      This is a private function containing blocking code.
      Please consider using ``get_timestamp`` (without the `_`),
      to schedule as a task in the event loop.

    """
    try:
        return hexlify(
            pack(STRUCT_PACKING_FORMAT, int(round(time_time())))
        ).decode(ENCODING_CODEC)
    except HANDLED_EXCEPTIONS as ex:
        raise DecodingError("failed to generate timestamp") from ex


async def get_timestamp(loop: AbstractEventLoop) -> str:
    """Asynchronous wrapper for _get_timestamp.

    Use as async wrapper for calling _get_timestamp,
    creating hexadecimal represntation of the current timestamp.

    Return:
      Hexadecimal represntation of the current unix time retrieved by
      ``time.time``.

    """
    return await loop.run_in_executor(None, _get_timestamp)


def _create_weekdays_value(requested_days: List[int]) -> str:
    """Create hex value from list of requested days for schedule updating.

    Args:
      data: list of integers represnting the requested days.
            check ``aioswitcher.consts`` for the correct values.

    Return:
      Hexadecimal representation of the days list.

    Raises:
      aioswitcher.erros.EncodingError: when failed to convert the argument.

    Note:
      This is a private function containing blocking code.
      Please consider using ``create_weekdays_value``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        if requested_days:
            return "{:02x}".format(int(sum(requested_days)))
        raise ValueError("days list is empty")
    except HANDLED_EXCEPTIONS as ex:
        raise EncodingError("failed to create weekdays value") from ex


async def create_weekdays_value(
    loop: AbstractEventLoop, requested_days: List[int]
) -> str:
    """Asynchronous wrapper for _create_weekdays_value.

    Use as async wrapper for calling _create_weekdays_value,
    creating hex value from list of requested days for schedule updating.

    Args:
      data: list of integers represnting the requested days.
            check ``aioswitcher.consts`` for the correct values.

    Return:
      Hexadecimal representation of the days list.

    """
    return await loop.run_in_executor(
        None, _create_weekdays_value, requested_days
    )


def _timedelta_str_to_schedule_time(time_value: str) -> str:
    """Convert time string to schedule start/end time to hexadecimale.

    Args:
      data: time to convert. e.g. "21:00".

    Return:
      Hexadecimal representation of time_value argument.

    Raises:
      aioswitcher.erros.EncodingError: when failed to convert the argument.

    Note:
      This is a private function containing blocking code.
      Please consider using ``timedelta_str_to_schedule_time``
      (without the `_`), to schedule as a task in the event loop.

    """
    try:
        return_time = mktime(
            strptime(
                strftime("%d/%m/%Y")
                + " "
                + time_value.split(":")[0]
                + ":"
                + time_value.split(":")[1],
                "%d/%m/%Y %H:%M",
            )
        )

        return hexlify(pack(STRUCT_PACKING_FORMAT, int(return_time))).decode(
            ENCODING_CODEC
        )
    except HANDLED_EXCEPTIONS as ex:
        raise EncodingError(
            "failed to convert time value to schedule time"
        ) from ex


async def timedelta_str_to_schedule_time(
    loop: AbstractEventLoop, time_value: str
) -> str:
    """Asynchronous wrapper for _timedelta_str_to_schedule_time.

    Use as async wrapper for calling _timedelta_str_to_schedule_time,
    converting time string to schedule start/end time to hexadecimale.

    Args:
      data: time to convert. e.g. "21:00".

    Return:
      Hexadecimal representation of time_value argument.

    """
    return await loop.run_in_executor(
        None, _timedelta_str_to_schedule_time, time_value
    )



"""Switcher water heater unofficial API and bridge, Bridge Object.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

from asyncio import AbstractEventLoop, Event, Queue
from functools import partial
from socket import AF_INET
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Type

from ..consts import SOCKET_BIND_TUP
from ..protocols import SwitcherV2UdpProtocolFactory

if TYPE_CHECKING:
    from ..devices import SwitcherV2Device


class SwitcherV2Bridge:
    """Represntation of the SwitcherV2 Bridge object.

    Args:
      loop: the event loop for the factory to run in.
      phone_id: the phone id of the desired device.
      device_id: the id of the desired device.
      device_password: the password of the desired device.

    Todo:
      * replace ``queue`` attribute with ``get_queue`` method.

    """

    def __init__(
        self,
        loop: AbstractEventLoop,
        phone_id: str,
        device_id: str,
        device_password: str,
    ) -> None:
        """Initialize the switcherv2 bridge."""
        self._loop = loop
        self._phone_id = phone_id
        self._device_id = device_id
        self._device_password = device_password

        self._device = None  # type: Optional[SwitcherV2Device]
        self._running_evt = Event()
        self._queue = Queue(maxsize=1)  # type: Queue

    async def __aenter__(self) -> "SwitcherV2Bridge":
        """Enter SwitcherV2Bridge asynchronous context manager.

        Returns:
          This instance of ``aioswitcher.bridge.SwitcherV2Bridge`` as an
          awaitable.

        """
        await self.start()
        return await self.__await__()

    async def __await__(self) -> "SwitcherV2Bridge":
        """Return SwitcherV2Bridge awaitable object.

        Returns:
          This instance of ``aioswitcher.bridge.SwitcherV2Bridge``.

        """
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit the SwitcherV2Bridge asynchronous context manager."""
        return await self.stop()

    async def start(self) -> None:
        """Create an asynchronous listenr and start the bridge event."""
        await self._loop.create_datagram_endpoint(
            partial(
                SwitcherV2UdpProtocolFactory,
                *[
                    self._loop,
                    self._phone_id,
                    self._device_id,
                    self._device_password,
                    self.queue,
                    self._running_evt,
                ],
            ),
            local_addr=SOCKET_BIND_TUP,
            family=AF_INET,
        )

        self._running_evt.set()
        return None

    async def stop(self) -> None:
        """Stop the asynchronous bridge."""
        self._running_evt.clear()
        return None

    @property
    def running(self) -> bool:
        """bool: Return true if bridge is running."""
        return self._running_evt.is_set()

    @property
    def queue(self) -> Queue:
        """asyncio.Queue: Return the queue storing updated device objects."""
        return self._queue



"""Switcher water heater unofficial API and bridge, Schedules.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

# fmt: off
from asyncio import AbstractEventLoop, Future, ensure_future
from binascii import unhexlify
from datetime import datetime
from typing import List

from .consts import (ALL_DAYS, MONDAY, SCHEDULE_DUE_ANOTHER_DAY_FORMAT,
                     SCHEDULE_DUE_TODAY_FORMAT, SCHEDULE_DUE_TOMMOROW_FORMAT,
                     SUNDAY, WAITING_TEXT, WEEKDAY_TUP)
from .tools import get_days_list_from_bytes, get_time_from_bytes

# fmt: on


class SwitcherV2Schedule:
    """Represnation of the SwitcherV2 schedule slot.

    Args:
      loop: the event loop to perform schedule operation in.
      idx: the index of the schedule slot (0-7).
      schedule_details: the string raw schedule data details.

    Todo:
      * Replace ``init_future`` attribute with ``get_init_future`` method.

    """

    def __init__(
        self, loop: AbstractEventLoop, idx: int, schedule_details: List[bytes]
    ) -> None:
        """Initialize the schedule."""
        self._loop = loop
        self._enabled = False
        self._recurring = False
        self._schedule_id = str(int(schedule_details[idx][0:2], 16))
        self._days = []  # type: List[str]
        self._schedule_data = b""
        self._start_time = WAITING_TEXT
        self._end_time = WAITING_TEXT
        self._duration = WAITING_TEXT
        self._init_future = loop.create_future()
        ensure_future(self.initialize(idx, schedule_details), loop=loop)

    async def initialize(
        self, idx: int, schedule_details: List[bytes]
    ) -> None:
        """Finish the initialization of the schedule."""
        try:
            if int(schedule_details[idx][2:4], 16) == 1:
                self._enabled = True
            if not schedule_details[idx][4:6] == "00":
                self._recurring = True
                if schedule_details[idx][4:6] == "fe":
                    self._days.append(ALL_DAYS)
                else:
                    self._days = await get_days_list_from_bytes(
                        self._loop,
                        bytearray(unhexlify((schedule_details[idx][4:6])))[0],
                    )

            self._start_time = await get_time_from_bytes(
                self._loop, schedule_details[idx][8:16]
            )
            self._end_time = await get_time_from_bytes(
                self._loop, schedule_details[idx][16:24]
            )
            self._duration = str(
                datetime.strptime(self._end_time, "%H:%M")
                - datetime.strptime(self._start_time, "%H:%M")
            )

            time_id = schedule_details[idx][0:2]
            on_off = schedule_details[idx][2:4]
            week = schedule_details[idx][4:6]
            timestate = schedule_details[idx][6:8]
            start_time = schedule_details[idx][8:16]
            end_time = schedule_details[idx][16:24]
            self._schedule_data = (
                time_id + on_off + week + timestate + start_time + end_time
            )

            self.init_future.set_result(self)
        except RuntimeError as exc:
            self.init_future.set_exception(exc)

        return None

    @property
    def schedule_id(self) -> str:
        """str: Return the schedule id."""
        return self._schedule_id

    @property
    def enabled(self) -> bool:
        """bool: Return true if enabled, setter included."""
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        """Setter to set the device as enabled or disabled."""
        if isinstance(value, bool):
            self._enabled = value
        else:
            raise TypeError("enabled is of type bool")

    @property
    def recurring(self) -> bool:
        """bool: Return true if recurring."""
        return self._recurring

    @property
    def days(self) -> List[str]:
        """list(str): Return the weekdays of the schedule."""
        return self._days

    @property
    def start_time(self) -> str:
        """str: Return the start time of the schedule."""
        return self._start_time

    @property
    def end_time(self) -> str:
        """str: Return the end time of the schedule."""
        return self._end_time

    @property
    def duration(self) -> str:
        """str: Return the duration of the schedule."""
        return self._duration

    @property
    def schedule_data(self) -> bytes:
        """bytes: Return the schedule data, setter included."""
        return self._schedule_data

    @schedule_data.setter
    def schedule_data(self, data: bytes) -> None:
        """Setter to set the schedule data for managing the schedule."""
        if isinstance(data, bytes):
            self._schedule_data = data
        else:
            raise TypeError("schedeule_data is of type bytes")

    @property
    def init_future(self) -> Future:
        """asyncio.Future: Return the future of the initialization."""
        return self._init_future

    def as_dict(self):
        """Return as dict.

        Returns:
          A dictionary represntation of the object properties.
          Used to make the object json serializable.

        """
        return self.__dict__


def _calc_next_run_for_schedule(schedule_details: SwitcherV2Schedule) -> str:
    """Calculate the next runtime of the schedule.

    Args:
      schedule_details: ``SwitcherV2Schedule`` representing the schedule slot.

    Returns:
      A pretty string describing the next due run.
      e.g. "Due tommorow at 17:00".

    Note:
      This is a private function containing blocking code.
      Please consider using ``calc_next_run_for_schedule`` (without the `_`),
      to schedule as a task on the event loop.

    """
    if schedule_details.recurring:
        today_datetime = datetime.now()

        start_time = datetime.strptime(schedule_details.start_time, "%H:%M")
        current_time = datetime.strptime(
            ("0" + str(today_datetime.hour))[-2:]
            + ":"
            + ("0" + str(today_datetime.minute))[-2:],
            "%H:%M",
        )

        current_weekday = today_datetime.weekday()
        found_day = -1
        if schedule_details.days == [ALL_DAYS]:
            if current_time < start_time:
                return SCHEDULE_DUE_TODAY_FORMAT.format(
                    schedule_details.start_time
                )
            return SCHEDULE_DUE_TOMMOROW_FORMAT.format(
                schedule_details.start_time
            )

        for day in schedule_details.days:
            set_weekday = WEEKDAY_TUP.index(day)

            if set_weekday == current_weekday and current_time < start_time:
                return SCHEDULE_DUE_TODAY_FORMAT.format(
                    schedule_details.start_time
                )

            if found_day == -1 or found_day > set_weekday:
                found_day = set_weekday

        if found_day - 1 == current_weekday or (
            found_day == WEEKDAY_TUP.index(MONDAY)
            and current_weekday == WEEKDAY_TUP.index(SUNDAY)
        ):

            return SCHEDULE_DUE_TOMMOROW_FORMAT.format(
                schedule_details.start_time
            )

        return SCHEDULE_DUE_ANOTHER_DAY_FORMAT.format(
            WEEKDAY_TUP[found_day], schedule_details.start_time
        )

    return SCHEDULE_DUE_TODAY_FORMAT.format(schedule_details.start_time)


async def calc_next_run_for_schedule(
    loop: AbstractEventLoop, schedule_details: SwitcherV2Schedule
) -> str:
    """Asynchronous wrapper for _calc_next_run_for_schedule.

    Use as async wrapper for calling _calc_next_run_for_schedule,
    calculating the next runtime of the schedule.

    Args:
      loop: the event loop to execute the function in.
      schedule_details: ``SwitcherV2Schedule`` representing the schedule slot.

    Returns:
      A pretty string describing the next due run.
      e.g. "Due tommorow at 17:00".

    """
    return await loop.run_in_executor(
        None, _calc_next_run_for_schedule, schedule_details
    )



"""Switcher water heater unofficial API and bridge.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

name = "aioswitcher"

__all__ = ["api", "bridge", "devices", "schedules", "tools"]



"""Switcher water heater unofficial API and bridge, API Object.

.. codeauthor:: Tomer Figenblat <tomer.figenblat@gmail.com>

"""

# fmt: off
from asyncio import AbstractEventLoop, Event, open_connection, wait
from binascii import unhexlify
from datetime import timedelta
from socket import AF_INET
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Tuple, Type

from ..consts import NO_TIMER_REQUESTED, REMOTE_SESSION_ID, SOCKET_PORT
from ..tools import (convert_minutes_to_timer, convert_string_to_device_name,
                     convert_timedelta_to_auto_off,
                     crc_sign_full_packet_com_key, get_timestamp)
from . import messages, packets

if TYPE_CHECKING:
    from asyncio import StreamReader, StreamWriter

# fmt: on


class SwitcherV2Api:
    """Represntation of the SwitcherV2 API object.

    Args:
      loop: the event loop for the factory to run in.
      ip_addr: the ip address assigned to the device
      phone_id: the phone id of the desired device.
      device_id: the id of the desired device.
      device_password: the password of the desired device.

    Todo:
      * ``control_device`` takes a timer value that must be converted to hex
        before calling this method using the deisgnated tool. On the other
        hand, the rest of the action methods` takes their arguments raw and
        use the appropriate tool for converting themselves.
        This is confusing and needs to be adjusted.

    """

    def __init__(
        self,
        loop: AbstractEventLoop,
        ip_addr: str,
        phone_id: str,
        device_id: str,
        device_password: str,
    ) -> None:
        """Initialize the Switcher V2 API."""
        self._loop = loop
        self._ip_addr = ip_addr
        self._phone_id = phone_id
        self._device_id = device_id
        self._device_password = device_password
        self._reader = None  # type: Optional[StreamReader]
        self._writer = None  # type: Optional[StreamWriter]
        self._connected_evt = Event()

    async def __aenter__(self) -> "SwitcherV2Api":
        """Enter SwitcherV2Api asynchronous context manager.

        Returns:
          This instance of ``aioswitcher.api.SwitcherV2Api`` as an awaitable.

        """
        await self.connect()
        return await self.__await__()

    async def __await__(self) -> "SwitcherV2Api":
        """Return SwitcherV2Api awaitable object.

        Returns:
          This instance of ``aioswitcher.api.SwitcherV2Api``.

        """
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        """Exit SwitcherV2Api asynchronous context manager."""
        return await self.disconnect()

    async def connect(self) -> None:
        """Connect to asynchronous socket and get reader and writer object."""
        self._reader, self._writer = await open_connection(
            host=self._ip_addr,
            port=SOCKET_PORT,
            loop=self._loop,
            family=AF_INET,
        )

        self._connected_evt.set()
        return None

    async def disconnect(self) -> None:
        """Disconnect from asynchronous socket."""
        if self._writer:
            self._writer.close()
        self._connected_evt.clear()
        return None

    async def _full_login(
        self,
    ) -> Tuple[str, Optional[messages.SwitcherV2LoginResponseMSG]]:
        """Use for sending the login packet to the device.

        Returns:
          A tuple of two
          ``Tuple[str, Optional[messages.SwitcherV2LoginResponseMSG]]``.

          The first object is a string contianing the hexadecimal
          representation of the current unix timestamp.

          The second object will be An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2LoginResponseMSG``.

        Note:
          This is a private function, please consider using its wrapper
          function ``login`` instead.

        """
        timestamp = await get_timestamp(self._loop)

        packet = packets.LOGIN_PACKET.format(
            REMOTE_SESSION_ID, timestamp, self._phone_id, self._device_password
        )

        signed_packet = await crc_sign_full_packet_com_key(self._loop, packet)

        if self._writer and self._reader:
            self._writer.write(unhexlify(signed_packet))

            response = await self._reader.read(1024)

            return (
                timestamp,
                messages.SwitcherV2LoginResponseMSG(self._loop, response),
            )

        return timestamp, None

    async def login(self) -> Optional[messages.SwitcherV2LoginResponseMSG]:
        """Use as wrapper for sending the login packet to the device.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2LoginResponseMSG``.

        """
        full_login_tuple = await self._full_login()
        return full_login_tuple[1]

    async def _full_get_state(
        self,
    ) -> Tuple[
        str,
        Optional[messages.SwitcherV2LoginResponseMSG],
        Optional[messages.SwitcherV2StateResponseMSG],
    ]:
        """Use for sending the get state packet to the device.

        Returns:
          A tuple of three
          ``Tuple[str, Optional[messages.SwitcherV2LoginResponseMSG],
          Optional[messages.SwitcherV2StateResponseMSG]]``.

          The first object is a string contianing the hexadecimal
          representation of the current unix timestamp.

          The second object will be An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2LoginResponseMSG``.

          The third object will be An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2StateResponseMSG``.

        Note:
          This is a private function, please consider using its wrapper
          function ``get_state`` instead.

        """
        timestamp, login_response = await self._full_login()

        if login_response and login_response.successful:
            packet = packets.GET_STATE_PACKET.format(
                login_response.session_id, timestamp, self._device_id
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                state_response = messages.SwitcherV2StateResponseMSG(
                    self._loop, response
                )
                await wait([state_response.init_future])
                return (timestamp, login_response, state_response)

        return timestamp, login_response, None

    async def get_state(self) -> Optional[messages.SwitcherV2StateResponseMSG]:
        """Use as wrapper for sending the get state packet to the device.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2StateResponseMSG``.

        """
        full_state_tuple = await self._full_get_state()
        return full_state_tuple[2]

    async def control_device(
        self, command: str, timer: Optional[str] = None
    ) -> Optional[messages.SwitcherV2ControlResponseMSG]:
        """Use for sending the control packet to the device.

        Args:
          command: specify ``aioswitcher.consts.COMMAND_ON`` or
            ``aioswitcher.consts.COMMAND_OFF``.
          timer: if turning-on optionally incorporate a auto-off timer. Use
            ``aioswitcher.tools.convert_minutes_to_timer`` to create the
            appropriate timer value.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2ControlResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):

            if timer:
                minutes_timer = await convert_minutes_to_timer(
                    self._loop, timer
                )
            else:
                minutes_timer = NO_TIMER_REQUESTED

            packet = packets.SEND_CONTROL_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
                command,
                minutes_timer,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2ControlResponseMSG(
                    self._loop, response
                )

        return None

    async def set_auto_shutdown(
        self, full_time: timedelta
    ) -> Optional[messages.SwitcherV2SetAutoOffResponseMSG]:
        """Use for sending the set auto-off packet to the device.

        Args:
          full_time: timedelta value containg the configuration value for
            auto-shutdown. Accepts anythin between 1 and 3 hours.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2SetAutoOffResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):
            auto_off = await convert_timedelta_to_auto_off(
                self._loop, full_time
            )

            packet = packets.SET_AUTO_OFF_SET_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
                auto_off,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2SetAutoOffResponseMSG(
                    self._loop, response
                )

        return None

    async def set_device_name(
        self, name: str
    ) -> Optional[messages.SwitcherV2UpdateNameResponseMSG]:
        """Use for sending the set name packet to the device.

        Args:
          name: string name with the length of 2 >= x >= 32.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2UpdateNameResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):
            device_name = await convert_string_to_device_name(self._loop, name)

            packet = packets.UPDATE_DEVICE_NAME_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
                device_name,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2UpdateNameResponseMSG(
                    self._loop, response
                )

        return None

    async def get_schedules(
        self,
    ) -> Optional[messages.SwitcherV2GetScheduleResponseMSG]:
        """Use for retrival of the schedules from the device.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2GetScheduleResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):
            packet = packets.GET_SCHEDULES_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2GetScheduleResponseMSG(
                    self._loop, response
                )

        return None

    async def disable_enable_schedule(
        self, schedule_data: str
    ) -> Optional[messages.SwitcherV2DisableEnableScheduleResponseMSG]:
        """Use for disabling or enabling a schedule on the device.

        Args:
          schedule_data: formatted data for updating the schedule, can be
            obtained from the ``aioswitcher.schedules.SwitcherV2Schedule``
            object or created using the format
            ``aioswitcher.consts.SCHEDULE_CREATE_DATA_FORMAT`` filled with
            three values: weekdays, start-time and end-time. Weekdays can be
            created using the tool ``aioswitcher.tools.create_weekdays_value``,
            the start and end times can be created using the
            ``aioswitcher.tools.timedelta_str_to_schedule_time``

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2DisableEnableScheduleResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):
            packet = packets.DISABLE_ENABLE_SCHEDULE_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
                schedule_data,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2DisableEnableScheduleResponseMSG(
                    self._loop, response
                )

        return None

    async def delete_schedule(
        self, schedule_id: str
    ) -> Optional[messages.SwitcherV2DeleteScheduleResponseMSG]:
        """Use for deleting a schedule from the device.

        Args:
          schedule_id: the id of the schedule slot from the device, can be 0-7
            as there are 8 slots available. Can be obtained from the
            ``aioswitcher.schedules.SwitcherV2Schedule`` object.

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2DeleteScheduleResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):
            packet = packets.DELETE_SCHEDULE_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
                schedule_id,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2DeleteScheduleResponseMSG(
                    self._loop, response
                )

        return None

    async def create_schedule(
        self, schedule_data: str
    ) -> Optional[messages.SwitcherV2CreateScheduleResponseMSG]:
        """Use for creating a new schedule in the next empty schedule slot.

        Args:
          schedule_data: formatted data for updating the schedule, can be
            created using the format
            ``aioswitcher.consts.SCHEDULE_CREATE_DATA_FORMAT`` filled with
            three values: weekdays, start-time and end-time. Weekdays can be
            created using the tool ``aioswitcher.tools.create_weekdays_value``,
            the start and end times can be created using the
            ``aioswitcher.tools.timedelta_str_to_schedule_time``

        Returns:
          An instance of the serialized object
          ``aioswitcher.api.messages.SwitcherV2CreateScheduleResponseMSG``.

        """
        (
            timestamp,
            login_response,
            get_state_response,
        ) = await self._full_get_state()

        if (
            login_response
            and login_response.successful
            and get_state_response
            and get_state_response.successful
        ):
            packet = packets.CREATE_SCHEDULE_PACKET.format(
                login_response.session_id,
                timestamp,
                self._device_id,
                self._phone_id,
                self._device_password,
                schedule_data,
            )

            signed_packet = await crc_sign_full_packet_com_key(
                self._loop, packet
            )

            if self._writer and self._reader:
                self._writer.write(unhexlify(signed_packet))

                response = await self._reader.read(1024)

                return messages.SwitcherV2CreateScheduleResponseMSG(
                    self._loop, response
                )

        return None

    @property
    def connected(self) -> bool:
        """bool: Return true if api is connected."""
        return self._connected_evt.is_set()
